defmodule Chronoscope.Gemini.ConnectionClient do
  require Logger

  alias Chronoscope.Certificate
  alias Chronoscope.Gemini.Request
  alias Chronoscope.Gemini.Response

  @timeout_in_milliseconds 3000

  @ssl Application.compile_env(:chronoscope, :ssl, :ssl)

  def connect(%{host: host, port: port, path: _} = resource) do
    case ssl_connect(host, port) do
      {:ok, socket} -> make_request(socket, resource)
      {:error, {:tls_alert, {:handshake_failure, error}}} -> {:error, handshake_failure_message("#{error}")}
      {:error, :timeout} -> {:error, :timeout}
      {:error, error} -> {:error, inspect(error)}
      error -> {:error, inspect(error)}
    end
  end

  defp ssl_connect(host, port) do
    host
    |> String.to_charlist()
    |> @ssl.connect(port, tls_options(host), @timeout_in_milliseconds)
  end

  defp tls_options(host) do
    host
    |> :tls_certificate_check.options()
    |> Keyword.put(:verify_fun, {&verify_fun/3, [check_hostname: String.to_charlist(host)]})
  end

  defp verify_fun(certificate, {:bad_cert, :selfsigned_peer}, state) do
    :ssl_verify_hostname.verify_fun(certificate, :valid_peer, state)
  end

  defp verify_fun(certificate, event, state) do
    :ssl_verify_hostname.verify_fun(certificate, event, state)
  end

  defp make_request(socket, url) do
    :ok = @ssl.send(socket, Request.create(url))
    {:ok, peercert} = @ssl.peercert(socket)

    peercert
    |> await_response()
    |> tap(fn _ -> @ssl.close(socket) end)
  end

  defp await_response(peercert) do
    receive do
      {:ssl, _socket, response} ->
        parse_response(response, peercert)

      msg ->
        Logger.error("received unexpected message: #{inspect(msg)}")
        {:error, :no_response}
    after
      @timeout_in_milliseconds ->
        Logger.error("timed out waiting for response")
        {:error, :timeout}
    end
  end

  defp parse_response(response, peercert) do
    response
    |> Response.parse()

    case Response.parse(response) do
      {:ok, result} ->
        {:ok, Map.put(result, :cert_expiration, Certificate.expiration_date(peercert))}

      {:error, error} ->
        {:error, error}
    end
  end

  defp handshake_failure_message(error) do
    cond do
      error =~ ~r/\{bad_cert,hostname_check_failed\}$/ ->
        "The certificate is NOT trusted. The name in the certificate does not match the expected."

      error =~ ~r/\{bad_cert,unable_to_match_altnames\}$/ ->
        "The certificate is NOT trusted. The name in the certificate does not match the expected."

      true ->
        String.trim(error)
    end
  end
end
