defmodule ChronoscopeWeb.API.V1.NTS.KeyEstablishmentControllerTest do
  use ChronoscopeWeb.ConnCase, async: true

  alias Chronoscope.NTSMock

  import Mox

  setup :verify_on_exit!

  describe "GET /api/v1/nts/key-establishment" do
    test "requires a host name", %{conn: conn} do
      response =
        conn
        |> get(~p"/api/v1/nts/key-establishment")
        |> json_response(400)

      assert %{"error" => "missing host"} == response
    end

    test "uses the given host name", %{conn: conn} do
      NTSMock
      |> expect(:key_establishment, fn "localhost", 4460 -> {:ok, %{status: :ok}} end)

      response =
        conn
        |> get(~p"/api/v1/nts/key-establishment?host=localhost")
        |> json_response(200)

      assert %{"status" => "ok", "response" => %{"cookies" => 0}} == response
    end

    test "truncates the host name", %{conn: conn} do
      NTSMock
      |> expect(
        :key_establishment,
        fn "test.example.com.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456",
           4460 ->
          {:ok, %{status: :ok}}
        end
      )

      response =
        conn
        |> get(
          ~p"/api/v1/nts/key-establishment?host=test.example.com.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.0123456789"
        )
        |> json_response(200)

      assert %{"status" => "ok", "response" => %{"cookies" => 0}} == response
    end

    test "uses the given port number", %{conn: conn} do
      NTSMock
      |> expect(:key_establishment, fn "localhost", 4461 -> {:ok, %{status: :ok}} end)

      response =
        conn
        |> get(~p"/api/v1/nts/key-establishment?host=localhost&port=4461")
        |> json_response(200)

      assert %{"status" => "ok", "response" => %{"cookies" => 0}} == response
    end

    test "handles an out of range port number", %{conn: conn} do
      response =
        conn
        |> get(~p"/api/v1/nts/key-establishment?host=localhost&port=65536")
        |> json_response(400)

      assert %{"error" => "port out of range"} == response
    end

    test "handles a negative port number", %{conn: conn} do
      response =
        conn
        |> get(~p"/api/v1/nts/key-establishment?host=localhost&port=-4460")
        |> json_response(400)

      assert %{"error" => "port out of range"} == response
    end

    test "handles a bad port number", %{conn: conn} do
      response =
        conn
        |> get(~p"/api/v1/nts/key-establishment?host=localhost&port=AA60")
        |> json_response(400)

      assert %{"error" => "invalid port"} == response
    end

    test "returns an empty successful response", %{conn: conn} do
      NTSMock
      |> expect(:key_establishment, fn "localhost", 4460 -> {:ok, %{status: :ok}} end)

      response =
        conn
        |> get(~p"/api/v1/nts/key-establishment?host=localhost")
        |> json_response(200)

      assert %{"status" => "ok", "response" => %{"cookies" => 0}} == response
    end

    test "returns a full successful response", %{conn: conn} do
      NTSMock
      |> expect(:key_establishment, fn "localhost", 4460 ->
        {:ok, %{cookies: [[], [], []], cookie_length: 300}}
      end)

      response =
        conn
        |> get(~p"/api/v1/nts/key-establishment?host=localhost")
        |> json_response(200)

      assert %{"status" => "ok", "response" => %{"cookies" => 3, "cookie_length" => 300}} == response
    end
  end
end
